//+------------------------------------------------------------------+
//| 12:02 PM 5/31/2020                          common_utilities.mqh |
//|                          Copyright  2019-20, WHRoeder@yahoo.com |
//|                                        mailto:WHRoeder@yahoo.com |
//+------------------------------------------------------------------+
//{ Typedefs
#define  COUNT    uint           //  One based.
   #define  INV_COUNT      UINT_MAX
#define  INDEX    uint           // Zero based.
   #define  INV_INDEX      UINT_MAX
#define  LOTS     double
#define  MONEY    double
#define  PRICE    double         // A Price
#define     PRICE_MAX   PRICE( EMPTY_VALUE)
#define     PRICE_MIN   PRICE(-EMPTY_VALUE)     // Oil went negative in 3/2020.
#define     CHANGE      double   // Difference of two prices.
#define     PIPS        double   // Difference of two prices in units of PIPs.
#define     POINTS      int      // Difference of two prices in units of Points.
#define  SYMBOL   string
#define  TICKET   int
   #define  INV_TICKET  EMPTY
#define  MAX_DATETIME   D'3000.12.31 23:59:59'
#define  MIN_DATETIME   0  // Thu, 01 Jan 1970 00:00:00 GMT
#define  SECONDS  int   // Difference of two datetimes.
//} Typedefs
#define  SERIES(I)   (Bars - 1 - I) // As-series to non-series or back.
enum TradeOperation{          // Same as OP_xxx except a proper enumeration.
      eTO_BUY, eTO_SELL, eTO_BUYLIMIT, eTO_SELLLIMIT, eTO_BUYSTOP, eTO_SELLSTOP,
      eTO_BALANCE,                         ///< (6) May occur in history pool.
      eTO_CREDIT};                         ///< (7) May occur in history pool.
   const TradeOperation eTO_MKT_FIRST   = eTO_BUY;     ///< (0) OP_BUY.
   const TradeOperation eTO_MKT_LAST    = eTO_SELL;    ///< (1) OP_SELL.
   const TradeOperation eTO_PND_FIRST   = eTO_BUYLIMIT;///< (2) OP_BUYLIMIT.
   const TradeOperation eTO_PND_LAST    = eTO_SELLSTOP;///< (5) OP_SELLSTOP.
bool        is_open(TradeOperation aOp){     return   aOp >= eTO_MKT_FIRST
                                                   && aOp <= eTO_MKT_LAST;     }
bool        is_pending(TradeOperation aOp){  return   aOp >= eTO_MKT_FIRST
                                                   && aOp <= eTO_MKT_LAST;     }
//{ Boldable fonts.
enum Font{  eFONT_ARIAL,                  // Arial
            eFONT_ARIAL_BOLD,             // Arial Bold
            eFONT_ARIAL_BLACK,            // Arial Black
            eFONT_CONSOLAS,               // Consolas
            eFONT_CONSOLAS_BOLD,          // Consolas Bold
            eFONT_COURIER_NEW,            // Courier New
            eFONT_COURIER_NEW_BOLD,       // Courier New Bold
            eFONT_LUCIDA_CONSOLE,         // Lucida Console
            eFONT_TIMES_NEW_ROMAN,        // Times New Roman
            eFONT_TIMES_NEW_ROMAN_BOLD,   // Times New Roman Bold
            eFONT_VERDANA,                // Verdana
            eFONT_VERDANA_BOLD            // Verdana Bold
            };
string      as_string(Font aFont){
   static const string  skFontAsString[]={   "Arial",          "Arial Bold",
   "Arial Black",          "Consolas",       "Consolas Bold",  "Courier New",
   "Courier New Bold",     "Lucida Console", "Times New Roman",
   "Times New Roman Bold", "Verdana",        "Verdana Bold"    };
                                                return skFontAsString[aFont];  }
//} Boldable fonts.
//{ Synthetic Pip size.
COUNT    gPipDigits; // \ OnCalculate -> OnTickComputePip
CHANGE   gPipSize;   // / May change per tick.
void        OnTickComputePip(void){
   MONEY    tickValue   = MarketInfo(_Symbol, MODE_TICKVALUE);
   CHANGE   tickSize    = MarketInfo(_Symbol, MODE_TICKSIZE);
   if(tickValue * tickSize != 0){   // Handle metals and exotics USDMXN/USDZAR.
      MONEY    dvpl  = tickValue / tickSize;                // DeltaValuePerlot.
                                    gPipSize  = _Point;  gPipDigits = 0;
      while(dvpl * gPipSize < 5){   gPipSize *= 10;    ++gPipDigits; }
   }else{   // Fall back.
      gPipDigits  = Digits % 2;
      gPipSize    = _Point * (gPipDigits==0 ? 1 : 10);
   }
}
PIPS        points_to_pips(POINTS anPnt){ return change_to_pips(anPnt*_Point); }
PIPS        change_to_pips(CHANGE aDiff){ return aDiff / gPipSize;             }
CHANGE      pips_to_change(CHANGE anPips){return anPips * gPipSize;            }
string      pips_as_string(PIPS anPips, string aPlusSign=""){
   return as_signed_fixed(anPips, gPipDigits, aPlusSign);                      }
//} Synthetic Pip size.
template<typename T>
string      as_signed(T aT, string aPlusSign="+"){
   if(aT < 0)        aPlusSign   = "";
   return aPlusSign + string(aT);
}
string      as_signed_fixed(double aValue, int anDecimals,string aPlusSign="+"){
   if(aValue < 0.)   aPlusSign   = "";
   return aPlusSign + DoubleToStr(aValue, anDecimals);
}
void        alert_id(string aMsg){
	Alert(StringFormat("%s %s %s %s", 	WindowExpertName(), _Symbol,
													as_string(PERIOD_CURRENT), aMsg) );
   }
void        print_id(string aMsg){
   PrintFormat(       "%s %s %s %s", 	WindowExpertName(), _Symbol,
                                             as_string(PERIOD_CURRENT), aMsg);
}
string      as_string(ENUM_TIMEFRAMES aePeriod){
   if(aePeriod == PERIOD_CURRENT)   aePeriod = ENUM_TIMEFRAMES(_Period);
   string   period_xxx  = EnumToString(aePeriod);                 // PERIOD_XXX
   return StringSubstr(period_xxx, 7);                            // XXX
}
string      as_string(ENUM_APPLIED_PRICE aePrice, int anLength=0){
   string price_XXX = EnumToString(aePrice);                      // PRICE_XXX
   return StringSubstr(price_XXX, 6, anLength);                   // XXX
}
string      as_string(ENUM_MA_METHOD aeMethod, int anLength=0){
   string mode_XXX = EnumToString(aeMethod);                      // MODE_XXX
   return StringSubstr(mode_XXX, 5, anLength);                    // XXX
}

bool        download_history(ENUM_TIMEFRAMES aePeriod=PERIOD_CURRENT){
   return download_history(_Symbol, aePeriod);
}
bool        download_history( SYMBOL            aSymbol,
                              ENUM_TIMEFRAMES   aePeriod=PERIOD_CURRENT){
   ResetLastError();    datetime other = iTime(aSymbol, aePeriod, 0);
   if(_LastError == 0 && other != 0)   return true;
   if(_LastError != ERR_HISTORY_WILL_UPDATED    // 4073
   && _LastError != ERR_NO_HISTORY_DATA         // 4066
     )   print_id( StringFormat("iTime(%s,%s) Failed: %i",
                                 aSymbol, as_string(aePeriod), _LastError) );
   return false;
}
template <typename T>
T           MathMax(T aA, T aB, T aC){ return MathMax(aA, MathMax(aB, aC) );   }
template <typename T>
T           MathMin(T aA, T aB, T aC){ return MathMin(aA, MathMin(aB, aC) );   }
template<typename T>
void        swap(T& aA, T& aB){        T t = aA;   aA = aB;    aB = t;         }

double      EMAl(double aPrevEMA, double aValue, double anLength){
   return EMAa( aPrevEMA, aValue, 2.0 / (anLength + 1.0) );
}
double      EMAa(double aPrevEMA, double aValue, double aAlpha){
   return aPrevEMA + aAlpha * (aValue - aPrevEMA);
}
double      PMAl(double aPrevPMA, double aValue, double anLength,double aPower){
   return PMAa(aPrevPMA, aValue, 2.0 / (anLength + 1.0), aPower);
}
double      PMAa(double aPrevPMA, double aValue, double aAlpha, double aPower){
   // en.wikipedia.org/wiki/Generalized_mean#Special_cases (Power Mean)
   aPrevPMA = MathPow(aPrevPMA, aPower);  aValue   = MathPow(aValue, aPower);
      double   ema   = EMAa(aPrevPMA, aValue, aAlpha);
   return   MathPow(ema, 1.0 / aPower);
}
double      true_range(INDEX aiBar){   PRICE    prevClose   = Close[aiBar+1];
   return   MathMax(High[aiBar], prevClose) - MathMin(Low[aiBar], prevClose);
}

double      MathNearest(  double aV, double aM){ return aM * MathRound(aV/aM); }
double      MathRoundDown(double aV, double aM){ return aM * MathFloor(aV/aM); }
double      MathRoundUp(  double aV, double aM){ return aM * MathCeil( aV/aM); }

enum TerminalMode{   eTM_DEBUG, eTM_LIVE, eTM_VISUAL, eTM_OPTIMIZER};
TerminalMode   get_modus_operandi(void){
   if (IS_DEBUG_MODE)                     return eTM_DEBUG;    // debugging.
   if (!MQLInfoInteger(MQL_TESTER) )      return eTM_LIVE;     // IsTesting()
   if ( MQLInfoInteger(MQL_VISUAL_MODE) ) return eTM_VISUAL;   // IsVisualMode()
                                          return eTM_OPTIMIZER;
}
CHANGE      get_spread_max(){
   //{ Make sure indicator "spread" is attached.
      static const bool    ksUseHA     = false;       // N/A
      static const COUNT   ksnDispWin  = INV_COUNT;   // Hide text.
      static const string  ksName      = "spread";
   double   NU = iCustom(  _Symbol, _Period, ksName,
                           ksUseHA, ksnDispWin,
                           0, 0);
   //} Make sure indicator "spread" is attached.
      static const string  skNameMax   = _Symbol + "_SpreadMax";
   return   GlobalVariableGet(skNameMax);
}
//+------------------------------------------------------------------+
//| Heiken Ashi class                                                |
//+------------------------------------------------------------------+
class HeikenAshi{
 public:
               HeikenAshi(ENUM_TIMEFRAMES aeTF=PERIOD_CURRENT)          // Ctor
                : mSym(_Symbol), meTF(aeTF){
                  if(meTF == PERIOD_CURRENT) meTF  = ENUM_TIMEFRAMES(_Period);
               }
               HeikenAshi( SYMBOL aSymbol,                              // Ctor
                           ENUM_TIMEFRAMES aeTF=PERIOD_CURRENT)
                : mSym(aSymbol), meTF(aeTF){
                  if(meTF == PERIOD_CURRENT) meTF  = ENUM_TIMEFRAMES(_Period);
               }
 public:                                                             // Methods
   enum HAbuffer{ eHA_LOW_HIGH,  eHA_HIGH_LOW,  eHA_OPEN,   eHA_CLOSE,
                  eHA_BOTTOM,    eHA_TOP  };
   PRICE       get(HAbuffer aeBuffer, INDEX aiBar)       const{
      static const string  ksName   = "Heiken Ashi";
      if(aeBuffer < eHA_BOTTOM)   return  iCustom( mSym, meTF, ksName,
                                          // All four parameters are just colors
                                                   aeBuffer, aiBar);
      PRICE    lh = iCustom(mSym, meTF, ksName, eHA_LOW_HIGH, aiBar);
      PRICE    hl = iCustom(mSym, meTF, ksName, eHA_HIGH_LOW, aiBar);
      return   aeBuffer == eHA_BOTTOM  ? MathMin(lh, hl)    : MathMax(lh, hl);
   }
   PRICE       get_mtf(HAbuffer aeBuffer, INDEX aiBar)   const{return
                                             get(aeBuffer, shift_mtf(aiBar));  }
   INDEX       shift_mtf(INDEX iBar)                     const{return
                                             iBarShift(mSym, meTF, Time[iBar]);}
 private:                                                            // Data
   SYMBOL            mSym;       // Chart symbol.
   ENUM_TIMEFRAMES   meTF;       // Chart time frame.
}; // HeikenAshi

//+------------------------------------------------------------------+
//| Average True Range                                               |
//+------------------------------------------------------------------+
class StandardATR{
 public:
               StandardATR(COUNT anLength=14)                           //[cTor]
                : mSym(_Symbol),    meTF(ENUM_TIMEFRAMES(_Period)),
                                                   mnLen(anLength){}
               StandardATR(ENUM_TIMEFRAMES aeTF, COUNT anLength=14)     // cTor
                : mSym(_Symbol),    meTF(aeTF),    mnLen(anLength){}
               StandardATR(SYMBOL aSymbol,                              // cTor
                           ENUM_TIMEFRAMES aeTF=PERIOD_CURRENT,
                           COUNT anLength=14)
                : mSym(aSymbol),    meTF(aeTF),    mnLen(anLength){
                  if(meTF == PERIOD_CURRENT) meTF  = ENUM_TIMEFRAMES(_Period);
               }
 public:                                                             // Methods
   PRICE       get(INDEX aiBar)           const{return
                                                iATR(mSym, meTF, mnLen, aiBar);}
   PRICE       get_mtf(INDEX aiBar)       const{return   get(shift_mtf(aiBar));}
   INDEX       shift_mtf(INDEX aiBar)     const{return
                                          iBarShift(mSym, meTF, Time[aiBar]);  }
   StandardATR*
               set_length(COUNT anLength){      mnLen = anLength; return &this;}
 private:                                                            // Data
   SYMBOL               mSym;       // Chart symbol.
   ENUM_TIMEFRAMES      meTF;       // Chart time frame.
   COUNT                mnLen;      // Length.
}; // StandardATR

//+------------------------------------------------------------------+
//| Moving Average                                                   |
//+------------------------------------------------------------------+
class StandardMA{
 public:
               StandardMA( ENUM_MA_METHOD aeMethod=MODE_SMA,            // cTor
                           COUNT anLength=14,
                           ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                           int anShift=0)
                : mSym(_Symbol),    meAP(aePrice),    meMAM(aeMethod),
                  meTF(ENUM_TIMEFRAMES(_Period)),     mnLen(anLength),
                  mnShift(anShift){}
               StandardMA( ENUM_TIMEFRAMES aeTF,                        // cTor
                           ENUM_MA_METHOD aeMethod=MODE_SMA, COUNT anLength=14,
                           ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                           int anShift=0)
                : mSym(_Symbol),    meAP(aePrice),    meMAM(aeMethod),
                  meTF(aeTF),       mnLen(anLength),  mnShift(anShift){}
               StandardMA( SYMBOL aSymbol,                              // cTor
                           ENUM_TIMEFRAMES aeTF=PERIOD_CURRENT,
                           ENUM_MA_METHOD aeMethod=MODE_SMA, COUNT anLength=14,
                           ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                           int anShift=0)
                : mSym(aSymbol),    meAP(aePrice),    meMAM(aeMethod),
                  meTF(aeTF),       mnLen(anLength),  mnShift(anShift){
                  if(meTF == PERIOD_CURRENT) meTF  = ENUM_TIMEFRAMES(_Period);
               }
 public:                                                             // Methods
   PRICE       get(INDEX aiBar)           const{return
                           iMA(mSym, meTF, mnLen, mnShift, meMAM, meAP, aiBar);}
   PRICE       get_mtf(INDEX aiBar)       const{return   get(shift_mtf(aiBar));}
   StandardMA* set_length(COUNT anLength){      mnLen = anLength; return &this;}
   INDEX       shift_mtf(INDEX iBar)      const{return
                                             iBarShift(mSym, meTF, Time[iBar]);}
 private:                                                            // Data
   SYMBOL               mSym;       // Chart symbol.
   ENUM_APPLIED_PRICE   meAP;       // Applied price.
   ENUM_MA_METHOD       meMAM;      // MA method.
   ENUM_TIMEFRAMES      meTF;       // Chart time frame.
   COUNT                mnLen;      // Length.
   int                  mnShift;
}; // StandardMA

//+------------------------------------------------------------------+
//| StdDev                                                           |
//+------------------------------------------------------------------+
class StandardDeviation{
 public:
            StandardDeviation(ENUM_MA_METHOD aeMethod=MODE_SMA,         // cTor
                              COUNT anLength=14,
                              ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                              int anShift=0)
                : mSym(_Symbol),    meAP(aePrice),    meMAM(aeMethod),
                  meTF(ENUM_TIMEFRAMES(_Period)),     mnLen(anLength),
                  mnShift(anShift){}
            StandardDeviation(ENUM_TIMEFRAMES aeTF,                     // cTor
                              ENUM_MA_METHOD aeMethod=MODE_SMA,
                              COUNT anLength=14,
                              ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                              int anShift=0)
                : mSym(_Symbol),    meAP(aePrice),    meMAM(aeMethod),
                  meTF(aeTF),       mnLen(anLength),  mnShift(anShift){}
            StandardDeviation(SYMBOL aSymbol,                           // cTor
                              ENUM_TIMEFRAMES aeTF=PERIOD_CURRENT,
                              ENUM_MA_METHOD aeMethod=MODE_SMA,
                              COUNT anLength=14,
                              ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                              int anShift=0)
                : mSym(aSymbol),    meAP(aePrice),    meMAM(aeMethod),
                  meTF(aeTF),       mnLen(anLength),  mnShift(anShift){
                  if(meTF == PERIOD_CURRENT) meTF  = ENUM_TIMEFRAMES(_Period);
               }
 public:                                                             // Methods
   PRICE       get(INDEX aiBar)           const{return
                     iStdDev(mSym, meTF, mnLen, mnShift, meMAM, meAP, aiBar);  }
   PRICE       get_mtf(INDEX aiBar)       const{return   get(shift_mtf(aiBar));}
   StandardDeviation*
               set_length(COUNT anLength){      mnLen = anLength; return &this;}
   INDEX       shift_mtf(INDEX iBar)      const{return
                                             iBarShift(mSym, meTF, Time[iBar]);}
 private:                                                            // Data
   SYMBOL               mSym;       // Chart symbol.
   ENUM_APPLIED_PRICE   meAP;       // Applied price.
   ENUM_MA_METHOD       meMAM;      // MA method.
   ENUM_TIMEFRAMES      meTF;       // Chart time frame.
   COUNT                mnLen;      // Length.
   int                  mnShift;
}; // StandardDeviation

//+------------------------------------------------------------------+
//| BollingerBands                                                   |
//+------------------------------------------------------------------+
class BollingerBands{
 public:
               BollingerBands(COUNT anLength=14, double aBWmult=2.0,    // cTor
                              ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                              int anShift=0)
                : mBWM(aBWmult),    mSym(_Symbol),    meAP(aePrice),
                  meTF(ENUM_TIMEFRAMES(_Period)),     mnLen(anLength),
                  mnShift(anShift){}
               BollingerBands(ENUM_TIMEFRAMES aeTF, COUNT anLength=14,  // cTor
                              double aBWmult=2.0,
                              ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                              int anShift=0)
                : mBWM(aBWmult),    mSym(_Symbol),    meAP(aePrice),
                  meTF(aeTF),       mnLen(anLength),  mnShift(anShift){}
               BollingerBands(SYMBOL aSymbol, ENUM_TIMEFRAMES aeTF,     // cTor
                              COUNT anLength=14, double aBWmult=2.0,
                              ENUM_APPLIED_PRICE aePrice=PRICE_CLOSE,
                              int anShift=0)
                : mBWM(aBWmult),    mSym(_Symbol),    meAP(aePrice),
                  meTF(aeTF),       mnLen(anLength),  mnShift(anShift){
                  if(meTF == PERIOD_CURRENT) meTF  = ENUM_TIMEFRAMES(_Period);
               }
 public:                                                             // Methods
   // Proper enumeration, not ints.
   enum BBbuffer{ eBB_MA=MODE_MAIN, eBB_UP=MODE_UPPER, eBB_LO=MODE_LOWER   };
   PRICE       get(BBbuffer aeBuffer, INDEX aiBar)    const{return
               iBands(mSym, meTF, mnLen, mBWM, mnShift, meAP, aeBuffer, aiBar);}
   PRICE       get_mtf(BBbuffer aeBuffer, INDEX aiBar)const{return
                                                get(aeBuffer, shift_mtf(aiBar));
   }
   BollingerBands*
               set_length(COUNT anLength){      mnLen = anLength; return &this;}
   BollingerBands*
               set_bwm(double aBWmult){         mBWM  = aBWmult;  return &this;}
   INDEX    shift_mtf(INDEX iBar)         const{return 
                                             iBarShift(mSym, meTF, Time[iBar]);}
 private:                                                            // Data
   double               mBWM;       // Bandwidth multiplier.
   SYMBOL               mSym;       // Chart symbol.
   ENUM_APPLIED_PRICE   meAP;       // Applied price.
   ENUM_TIMEFRAMES      meTF;       // Chart time frame.
   COUNT                mnLen;      // Length.
   int                  mnShift;    // MA shift.
}; // BollingerBands

//+------------------------------------------------------------------+
//| Ticker                                                           |
//+------------------------------------------------------------------+
class Ticker{
// Try to start the timer in OnCalculate, as EventSetTimer sometimes fails in
// OnInit, on MT4, with ERR_INTERNAL_ERROR (4024) and ERR_CHART_NOREPLY (4030)
 public:
               Ticker(void)    : mTS(eTS_IDLE){}                        //[CTor]
 public:                                                             // Methods
   bool        enable(COUNT anSeconds=1){
      if(mTS != eTS_ACTIVE){
         if(!EventSetTimer(anSeconds) ){
            alert_id(StringFormat("EST(%i):%i", anSeconds, _LastError) );
            mTS   = eTS_ERROR;
            return false;
         }
         if(mTS == eTS_ERROR) alert_id("EST: Successful");
         mTS   = eTS_ACTIVE;
      }
      return true;
   }
   void        disable(void){             EventKillTimer();    mTS = eTS_IDLE; }
 private:                                                            // Data
   enum TickerStatus{   eTS_IDLE,   eTS_ERROR,  eTS_ACTIVE  };
   TickerStatus   mTS;
}; // class Ticker
Ticker   gsTicker;   // Singleton.

//+------------------------------------------------------------------+
//| Timer                                                            |
//+------------------------------------------------------------------+
class Timer{
   //{Typedef
   #define TICK_COUNT uint
      #define TICK_COUNT_MAX UINT_MAX
   //}Typedef
 public:                                                                //[cTor]
               Timer(void)     		 : myTickCount( GetTickCount() ){}
 public:                                                             // Methods
   void        reset(void){      		myTickCount = GetTickCount();           }
   double      get_seconds(void)	const{return	get_ms() / 1000.0;				 }
   TICK_COUNT  get_ms(void)		const{							// In milliseconds.
      // Counter is limited by the restrictions of the system timer. Time is
      // stored as an unsigned integer, so it's overfilled every 49.7 days if a
      // computer works uninterruptedly.
      long     delay = (long)GetTickCount() - (long)myTickCount;
      if (delay < 0) delay += long(TICK_COUNT_MAX) + 1;     // GTC rolled over.
      return (TICK_COUNT)delay;
   }
 private:                                                            // Data
   TICK_COUNT  myTickCount;
}; // Timer
//+- common_utilities.mqh -------------------------------------------+
